@@ -6,28 +6,46 @@ from decimal import Decimal
 from dateutil import relativedelta
 from dateutil import rrule
 from itertools import groupby
-from cached_property import cached_property
 
+from trytond.i18n import gettext
 from trytond.model import Workflow, ModelSQL, ModelView, fields, Unique
+from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval, Bool, If
 from trytond.pool import Pool
+from trytond.tools import cached_property
 from trytond.transaction import Transaction
 from trytond.wizard import (Wizard, StateView, StateTransition, StateReport,
     Button)
 from trytond.tools import grouped_slice
 from trytond.modules.company import CompanyReport
 
-__all__ = ['Asset', 'AssetLine', 'AssetUpdateMove',
-    'CreateMovesStart', 'CreateMoves',
-    'UpdateAssetStart', 'UpdateAssetShowDepreciation', 'UpdateAsset',
-    'AssetDepreciationTable',
-    'PrintDepreciationTableStart', 'PrintDepreciationTable']
+from .exceptions import PrintDepreciationTableError
 
 
 def date2datetime(date):
     return datetime.datetime.combine(date, datetime.time())
 
 
+February = 2
+
+
+def normalized_delta(start, end):
+    "Returns timedelta using fixed 365 days per year"
+    assert start <= end
+    delta = end - start
+    correction = 0
+    if start.year == end.year:
+        if calendar.isleap(start.year) and start.month <= 2 and end.month > 2:
+            correction -= 1
+    else:
+        if calendar.isleap(start.year) and start.month <= February:
+            correction -= 1
+        if calendar.isleap(end.year) and end.month > February:
+            correction -= 1
+        correction -= calendar.leapdays(start.year + 1, end.year)
+    return delta + datetime.timedelta(days=correction)
+
+
 class Asset(Workflow, ModelSQL, ModelView):
     'Asset'
     __name__ = 'account.asset'
@@ -37,7 +55,10 @@ class Asset(Workflow, ModelSQL, ModelView):
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['state'],
+        context={
+            'company': Eval('company', None),
+            },
+        depends=['state', 'company'],
         domain=[
             ('type', '=', 'assets'),
             ('depreciable', '=', True),
@@ -45,7 +66,7 @@ class Asset(Workflow, ModelSQL, ModelView):
     supplier_invoice_line = fields.Many2One('account.invoice.line',
         'Supplier Invoice Line',
         domain=[
-            If(Eval('product', None) == None,
+            If(~Eval('product', None),
                 ('product', '=', -1),
                 ('product', '=', Eval('product', -1)),
                 ),
@@ -75,6 +96,8 @@ class Asset(Workflow, ModelSQL, ModelView):
             },
         depends=['state'],
         required=True)
+    currency = fields.Function(fields.Many2One('currency.currency',
+        'Currency'), 'on_change_with_currency')
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
         'on_change_with_currency_digits')
     quantity = fields.Float('Quantity',
@@ -99,12 +122,33 @@ class Asset(Workflow, ModelSQL, ModelView):
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
         depends=['currency_digits', 'state'],
-        required=True)
-    residual_value = fields.Numeric('Residual Value',
+        required=True,
+        help="The value of the asset when purchased.")
+    depreciated_amount = fields.Numeric("Depreciated Amount",
+        digits=(16, Eval('currency_digits', 2)),
+        domain=[
+            ('depreciated_amount', '<=', Eval('value')),
+            ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['currency_digits', 'state'],
+        depends=['currency_digits', 'value', 'state'],
+        required=True,
+        help="The amount already depreciated at the start date.")
+    depreciating_value = fields.Function(fields.Numeric(
+            "Depreciating Value",
+            digits=(16, Eval('currency_digits', 2)),
+            depends=['currency_digits'],
+            help="The value of the asset at the start date."),
+        'on_change_with_depreciating_value')
+    residual_value = fields.Numeric('Residual Value',
+        domain=[
+            ('residual_value', '<=', Eval('depreciating_value')),
+            ],
+        states={
+            'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
+            },
+        depends=['currency_digits', 'depreciating_value', 'state'],
         required=True,
         digits=(16, Eval('currency_digits', 2)))
     purchase_date = fields.Date('Purchase Date', states={
@@ -173,11 +217,8 @@ class Asset(Workflow, ModelSQL, ModelView):
         table = cls.__table__()
         cls._sql_constraints = [
             ('invoice_line_uniq', Unique(table, table.supplier_invoice_line),
-                'Supplier Invoice Line can be used only once on asset.'),
+                'account_asset.msg_asset_invoice_line_unique'),
             ]
-        cls._error_messages.update({
-                'delete_draft': 'Asset "%s" must be in draft to be deleted.',
-                })
         cls._transitions |= set((
                 ('draft', 'running'),
                 ('running', 'closed'),
@@ -220,14 +261,24 @@ class Asset(Workflow, ModelSQL, ModelView):
     def default_state():
         return 'draft'
 
-    @staticmethod
-    def default_frequency():
-        return 'monthly'
+    @classmethod
+    def default_frequency(cls, **pattern):
+        pool = Pool()
+        Configuration = pool.get('account.configuration')
+        return Configuration(1).get_multivalue('asset_frequency', **pattern)
 
     @staticmethod
     def default_depreciation_method():
         return 'linear'
 
+    @classmethod
+    def default_depreciated_amount(cls):
+        return Decimal(0)
+
+    @classmethod
+    def default_residual_value(cls):
+        return Decimal(0)
+
     @staticmethod
     def default_start_date():
         return Pool().get('ir.date').today()
@@ -236,6 +287,11 @@ class Asset(Workflow, ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    @fields.depends('company')
+    def on_change_company(self):
+        self.frequency = self.default_frequency(
+            company=self.company.id if self.company else None)
+
     @staticmethod
     def default_account_journal():
         Journal = Pool().get('account.journal')
@@ -246,6 +302,18 @@ class Asset(Workflow, ModelSQL, ModelView):
             return journals[0].id
         return None
 
+    @fields.depends('value', 'depreciated_amount')
+    def on_change_with_depreciating_value(self, name=None):
+        if self.value is not None and self.depreciated_amount is not None:
+            return self.value - self.depreciated_amount
+        else:
+            return Decimal(0)
+
+    @fields.depends('company')
+    def on_change_with_currency(self, name=None):
+        if self.company:
+            return self.company.currency.id
+
     @fields.depends('company')
     def on_change_with_currency_digits(self, name=None):
         if self.company:
@@ -321,9 +389,9 @@ class Asset(Workflow, ModelSQL, ModelView):
         return result
 
     def get_depreciated_amount(self):
-        lines = [line.accumulated_depreciation for line in self.lines
+        lines = [line.depreciation for line in self.lines
             if line.move and line.move.state == 'posted']
-        return max(lines) if lines else 0
+        return sum(lines, Decimal(0))
 
     def compute_move_dates(self):
         """
@@ -342,42 +410,52 @@ class Asset(Workflow, ModelSQL, ModelView):
             return [self.end_date]
         if self.frequency == 'monthly':
             rule = rrule.rrule(rrule.MONTHLY, dtstart=self.start_date,
-                bymonthday=int(config.asset_bymonthday))
+                bymonthday=int(config.get_multivalue(
+                        'asset_bymonthday', company=self.company.id)))
         elif self.frequency == 'yearly':
             rule = rrule.rrule(rrule.YEARLY, dtstart=self.start_date,
-                bymonth=int(config.asset_bymonth),
-                bymonthday=int(config.asset_bymonthday))
+                bymonth=int(config.get_multivalue(
+                            'asset_bymonth', company=self.company.id)),
+                bymonthday=int(config.get_multivalue(
+                            'asset_bymonthday', company=self.company.id)))
         dates = [d.date()
             for d in rule.between(date2datetime(start_date),
                 date2datetime(self.end_date))]
         dates.append(self.end_date)
         return dates
 
-    def compute_depreciation(self, date, dates):
+    def compute_depreciation(self, amount, date, dates):
         """
         Returns the depreciation amount for an asset on a certain date.
         """
-        amount = (self.value - self.get_depreciated_amount()
-            - self.residual_value)
         if self.depreciation_method == 'linear':
             start_date = max([self.start_date
                     - relativedelta.relativedelta(days=1)]
                 + [l.date for l in self.lines])
-            first_delta = dates[0] - start_date
+            first_delta = normalized_delta(start_date, dates[0])
             if len(dates) > 1:
-                last_delta = dates[-1] - dates[-2]
+                last_delta = normalized_delta(dates[-2], dates[-1])
             else:
                 last_delta = first_delta
             if self.frequency == 'monthly':
                 _, first_ndays = calendar.monthrange(
                     dates[0].year, dates[0].month)
+                if (calendar.isleap(dates[0].year)
+                        and dates[0].month == February):
+                    first_ndays -= 1
                 _, last_ndays = calendar.monthrange(
                     dates[-1].year, dates[-1].month)
+                if (calendar.isleap(dates[-1].year)
+                        and dates[-1].month == February):
+                    last_ndays -= 1
             elif self.frequency == 'yearly':
-                first_ndays = 365 + calendar.isleap(dates[0].year)
-                last_ndays = 365 + calendar.isleap(dates[-1].year)
-            first_ratio = Decimal(first_delta.days) / Decimal(first_ndays)
-            last_ratio = Decimal(last_delta.days) / Decimal(last_ndays)
+                first_ndays = last_ndays = 365
+            first_ratio = (
+                Decimal(min(first_delta.days, first_ndays))
+                / Decimal(first_ndays))
+            last_ratio = (
+                Decimal(min(last_delta.days, last_ndays))
+                / Decimal(last_ndays))
             depreciation = amount / (
                 len(dates) - 2 + first_ratio + last_ratio)
             if date == dates[0]:
@@ -393,14 +471,17 @@ class Asset(Workflow, ModelSQL, ModelView):
         Line = Pool().get('account.asset.line')
         amounts = {}
         dates = self.compute_move_dates()
-        amount = (self.value - self.get_depreciated_amount()
+        depreciated_amount = self.get_depreciated_amount()
+        amount = (self.depreciating_value
+            - depreciated_amount
             - self.residual_value)
         if amount <= 0:
             return amounts
-        residual_value, acc_depreciation = amount, Decimal(0)
+        residual_value, acc_depreciation = (
+            amount, depreciated_amount + self.depreciated_amount)
         asset_line = None
         for date in dates:
-            depreciation = self.compute_depreciation(date, dates)
+            depreciation = self.compute_depreciation(amount, date, dates)
             amounts[date] = asset_line = Line(
                 acquired_value=self.value,
                 depreciable_basis=amount,
@@ -408,22 +489,20 @@ class Asset(Workflow, ModelSQL, ModelView):
             if depreciation > residual_value:
                 asset_line.depreciation = residual_value
                 asset_line.accumulated_depreciation = (
-                    self.get_depreciated_amount()
-                    + acc_depreciation + residual_value)
+                    acc_depreciation + residual_value)
                 break
             else:
                 residual_value -= depreciation
                 acc_depreciation += depreciation
                 asset_line.depreciation = depreciation
-                asset_line.accumulated_depreciation = (
-                    self.get_depreciated_amount() + acc_depreciation)
+                asset_line.accumulated_depreciation = acc_depreciation
         else:
             if residual_value > 0 and asset_line is not None:
                 asset_line.depreciation += residual_value
                 asset_line.accumulated_depreciation += residual_value
         for asset_line in amounts.values():
-            asset_line.actual_value = (self.value -
-                asset_line.accumulated_depreciation)
+            asset_line.actual_value = (self.value
+                - asset_line.accumulated_depreciation)
         return amounts
 
     @classmethod
@@ -514,7 +593,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         Line.save(lines)
         Move.post(moves)
 
-    def get_closing_move(self, account):
+    def get_closing_move(self, account, date=None):
         """
         Returns closing move values.
         """
@@ -524,7 +603,8 @@ class Asset(Workflow, ModelSQL, ModelView):
         Move = pool.get('account.move')
         MoveLine = pool.get('account.move.line')
 
-        date = Date.today()
+        if date is None:
+            date = Date.today()
         period_id = Period.find(self.company.id, date)
         if self.supplier_invoice_line:
             account_asset = self.supplier_invoice_line.account.current()
@@ -537,7 +617,7 @@ class Asset(Workflow, ModelSQL, ModelView):
             account=account_asset,
             )
         depreciation_line = MoveLine(
-            debit=self.get_depreciated_amount(),
+            debit=self.get_depreciated_amount() + self.depreciated_amount,
             credit=0,
             account=self.product.account_depreciation_used,
             )
@@ -567,14 +647,14 @@ class Asset(Workflow, ModelSQL, ModelView):
         Fill the number field with asset sequence.
         '''
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Config = pool.get('account.configuration')
 
         config = Config(1)
         for asset in assets:
             if asset.number:
                 continue
-            asset.number = Sequence.get_id(config.asset_sequence.id)
+            asset.number = config.get_multivalue(
+                'asset_sequence', company=asset.company.id).get()
         cls.save(assets)
 
     @classmethod
@@ -587,7 +667,7 @@ class Asset(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('closed')
-    def close(cls, assets, account=None):
+    def close(cls, assets, account=None, date=None):
         """
         Close the assets.
         If account is provided, it will be used instead of the expense account.
@@ -597,7 +677,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         cls.clear_lines(assets)
         moves = []
         for asset in assets:
-            moves.append(asset.get_closing_move(account))
+            moves.append(asset.get_closing_move(account, date=date))
         Move.save(moves)
         for move, asset in zip(moves, assets):
             asset.move = move
@@ -631,7 +711,9 @@ class Asset(Workflow, ModelSQL, ModelView):
     def delete(cls, assets):
         for asset in assets:
             if asset.state != 'draft':
-                cls.raise_user_error('delete_draft', asset.rec_name)
+                raise AccessError(
+                    gettext('account_asset.msg_delete_draft',
+                        asset=asset.rec_name))
         return super(Asset, cls).delete(assets)
 
 
@@ -642,20 +724,39 @@ class AssetLine(ModelSQL, ModelView):
         ondelete='CASCADE', readonly=True)
     date = fields.Date('Date', readonly=True)
     depreciation = fields.Numeric('Depreciation',
-        digits=(16, Eval('_parent_asset', {}).get('currency_digits', 2)),
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'],
         required=True, readonly=True)
-    acquired_value = fields.Numeric('Acquired Value', readonly=True)
-    depreciable_basis = fields.Numeric('Depreciable Basis', readonly=True)
-    actual_value = fields.Numeric('Actual Value', readonly=True)
+    acquired_value = fields.Numeric('Acquired Value', readonly=True,
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
+    depreciable_basis = fields.Numeric('Depreciable Basis', readonly=True,
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
+    actual_value = fields.Numeric('Actual Value', readonly=True,
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
     accumulated_depreciation = fields.Numeric(
-        'Accumulated Depreciation', readonly=True)
+        'Accumulated Depreciation', readonly=True,
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
     move = fields.Many2One('account.move', 'Account Move', readonly=True)
+    currency = fields.Function(fields.Many2One('currency.currency',
+        'Currency'), 'on_change_with_currency')
+    currency_digits = fields.Function(fields.Integer('Currency Digits'),
+        'on_change_with_currency_digits')
 
     @classmethod
     def __setup__(cls):
         super(AssetLine, cls).__setup__()
+        cls.__access__.add('asset')
         cls._order.insert(0, ('date', 'ASC'))
 
+    @fields.depends('asset', '_parent_asset.currency')
+    def on_change_with_currency(self, name=None):
+        if self.asset:
+            return self.asset.currency.id
+
+    @fields.depends('asset', '_parent_asset.currency_digits')
+    def on_change_with_currency_digits(self, name=None):
+        if self.asset:
+            return self.asset.currency_digits
+
 
 class AssetUpdateMove(ModelSQL):
     'Asset - Update - Move'
@@ -688,9 +789,11 @@ class CreateMoves(Wizard):
 
     def transition_create_moves(self):
         Asset = Pool().get('account.asset')
-        assets = Asset.search([
-                ('state', '=', 'running'),
-                ])
+        with Transaction().set_context(_check_access=True):
+            assets = Asset.search([
+                    ('state', '=', 'running'),
+                    ])
+        assets = Asset.browse(assets)
         Asset.create_moves(assets, self.start.date)
         return 'end'
 
@@ -709,8 +812,8 @@ class UpdateAssetShowDepreciation(ModelView):
     amount = fields.Numeric('Amount', readonly=True)
     date = fields.Date('Date', required=True,
         domain=[
-            ('date', '>', Eval('latest_move_date')),
-            ('date', '<', Eval('next_depreciation_date')),
+            ('date', '>=', Eval('latest_move_date')),
+            ('date', '<=', Eval('next_depreciation_date')),
             ],
         depends=['latest_move_date', 'next_depreciation_date'],
         help=('The date must be between the last update/depreciation date '
@@ -742,18 +845,14 @@ class UpdateAsset(Wizard):
     create_lines = StateTransition()
 
     def default_start(self, fields):
-        Asset = Pool().get('account.asset')
-        asset = Asset(Transaction().context['active_id'])
         return {
-            'value': asset.value,
-            'residual_value': asset.residual_value,
-            'end_date': asset.end_date,
+            'value': self.record.value,
+            'residual_value': self.record.residual_value,
+            'end_date': self.record.end_date,
             }
 
     def transition_update_asset(self):
-        Asset = Pool().get('account.asset')
-        asset = Asset(Transaction().context['active_id'])
-        if self.start.value != asset.value:
+        if self.start.value != self.record.value:
             return 'show_move'
         return 'create_lines'
 
@@ -773,15 +872,15 @@ class UpdateAsset(Wizard):
         return min(next_dates)
 
     def default_show_move(self, fields):
-        Asset = Pool().get('account.asset')
-        asset = Asset(Transaction().context['active_id'])
         return {
-            'amount': self.start.value - asset.value,
+            'amount': self.start.value - self.record.value,
             'date': datetime.date.today(),
-            'depreciation_account': asset.product.account_depreciation_used.id,
-            'counterpart_account': asset.product.account_expense_used.id,
-            'latest_move_date': self.get_latest_move_date(asset),
-            'next_depreciation_date': self.get_next_depreciation_date(asset),
+            'depreciation_account': (
+                self.record.product.account_depreciation_used.id),
+            'counterpart_account': self.record.product.account_expense_used.id,
+            'latest_move_date': self.get_latest_move_date(self.record),
+            'next_depreciation_date': self.get_next_depreciation_date(
+                self.record),
             }
 
     def get_move(self, asset):
@@ -814,32 +913,27 @@ class UpdateAsset(Wizard):
     def transition_create_move(self):
         pool = Pool()
         Move = pool.get('account.move')
-        Asset = pool.get('account.asset')
 
-        asset = Asset(Transaction().context['active_id'])
         latest_move_date = self.show_move.latest_move_date
         next_date = self.show_move.next_depreciation_date
-        if not (latest_move_date < self.show_move.date < next_date):
+        if not (latest_move_date <= self.show_move.date <= next_date):
             raise ValueError('The update move date is invalid')
-        move = self.get_move(asset)
-        move.lines = self.get_move_lines(asset)
+        move = self.get_move(self.record)
+        move.lines = self.get_move_lines(self.record)
         move.save()
-        Asset.write([asset], {
+        self.model.write([self.record], {
                 'update_moves': [('add', [move.id])],
                 })
         Move.post([move])
         return 'create_lines'
 
     def transition_create_lines(self):
-        Asset = Pool().get('account.asset')
-
-        asset = Asset(Transaction().context['active_id'])
-        Asset.write([asset], {
+        self.model.write([self.record], {
                 'value': self.start.value,
                 'residual_value': self.start.residual_value,
                 'end_date': self.start.end_date,
                 })
-        Asset.create_lines([asset])
+        self.model.create_lines([self.record])
         return 'end'
 
 
@@ -848,8 +942,8 @@ class AssetDepreciationTable(CompanyReport):
     __name__ = 'account.asset.depreciation_table'
 
     @classmethod
-    def get_context(cls, records, data):
-        context = super(AssetDepreciationTable, cls).get_context(records, data)
+    def get_context(cls, records, header, data):
+        context = super().get_context(records, header, data)
 
         AssetDepreciation = cls.get_asset_depreciation()
         AssetDepreciation.start_date = data['start_date']
@@ -895,7 +989,10 @@ class AssetDepreciationTable(CompanyReport):
             'closing_value',
             }
         for attr_name in grouped_attributes:
-            setattr(Grouper, attr_name, cached_property(adder(attr_name)))
+            descr = cached_property(adder(attr_name))
+            setattr(Grouper, attr_name, descr)
+            if hasattr(descr, '__set_name__'):
+                descr.__set_name__(Grouper, attr_name)
 
         return Grouper
 
@@ -913,14 +1010,18 @@ class AssetDepreciationTable(CompanyReport):
 
             @cached_property
             def update_lines(self):
-                filter_ = lambda l: (l.account.kind == 'other'
-                    and self.start_date < l.move.date <= self.end_date)
+                def filter_(l):
+                    return (l.account.type.expense
+                        and self.start_date < l.move.date <= self.end_date)
                 return list(filter(filter_,
                     (l for m in self.asset.update_moves for l in m.lines)))
 
             @cached_property
             def start_fixed_value(self):
-                if self.start_date < self.asset.start_date:
+                if self.asset.end_date < self.start_date:
+                    return self.asset.value
+                elif (self.start_date < self.asset.start_date
+                        or not self.asset_lines):
                     return 0
                 value = self.asset_lines[0].acquired_value
                 date = self.asset_lines[0].date
@@ -946,7 +1047,7 @@ class AssetDepreciationTable(CompanyReport):
             @cached_property
             def end_fixed_value(self):
                 if not self.asset_lines:
-                    return 0
+                    return self.start_fixed_value
                 value = self.asset_lines[-1].acquired_value
                 date = self.asset_lines[-1].date
                 for line in self.update_lines:
@@ -957,7 +1058,10 @@ class AssetDepreciationTable(CompanyReport):
             @cached_property
             def start_value(self):
                 if not self.asset_lines:
-                    return self.asset.value
+                    if self.asset.start_date > self.end_date:
+                        return self.asset.value
+                    else:
+                        return 0
                 return (self.asset_lines[0].actual_value
                     + self.asset_lines[0].depreciation)
 
@@ -974,7 +1078,7 @@ class AssetDepreciationTable(CompanyReport):
             @cached_property
             def end_value(self):
                 if not self.asset_lines:
-                    return self.asset.value
+                    return self.start_value
                 return self.asset_lines[-1].actual_value
 
             @cached_property
@@ -1030,11 +1134,21 @@ class PrintDepreciationTable(Wizard):
     def do_print_(self, action):
         pool = Pool()
         Asset = pool.get('account.asset')
-        assets = Asset.search([
-                ('start_date', '<', self.start.end_date),
-                ('end_date', '>', self.start.start_date),
-                ('state', '=', 'running'),
-                ])
+        with Transaction().set_context(_check_access=True):
+            assets = Asset.search([
+                    ['OR',
+                        ('purchase_date', '<=', self.start.end_date),
+                        ('start_date', '<=', self.start.end_date),
+                        ],
+                    ['OR',
+                        ('move', '=', None),
+                        ('move.date', '>=', self.start.start_date),
+                        ],
+                    ('state', '!=', 'draft'),
+                    ])
+        if not assets:
+            raise PrintDepreciationTableError(
+                gettext('account_asset.msg_no_assets'))
         return action, {
             'ids': [a.id for a in assets],
             'start_date': self.start.start_date,
